@@ -9,7 +9,7 @@ import (
 	"time"
 
 	"github.com/clarketm/json"
-	ign3types "github.com/coreos/ignition/v2/config/v3_1/types"
+	ign3types "github.com/coreos/ignition/v2/config/v3_2/types"
 	"github.com/golang/glog"
 	"github.com/imdario/mergo"
 	"github.com/vincent-petithory/dataurl"
@@ -88,10 +88,6 @@ type Controller struct {
 
 	queue        workqueue.RateLimitingInterface
 	featureQueue workqueue.RateLimitingInterface
-
-	// we need this method to mock out patch calls in unit until https://github.com/openshift/machine-config-operator/pull/611#issuecomment-481397185
-	// which is probably going to be in kube 1.14
-	patchKubeletConfigsFunc func(string, []byte) error
 }
 
 // New returns a new kubelet config controller
@@ -143,8 +139,6 @@ func New(
 	ctrl.featLister = featInformer.Lister()
 	ctrl.featListerSynced = featInformer.Informer().HasSynced
 
-	ctrl.patchKubeletConfigsFunc = ctrl.patchKubeletConfigs
-
 	return ctrl
 }
 
@@ -292,11 +286,6 @@ func (ctrl *Controller) handleErr(err error, key interface{}) {
 		return
 	}
 
-	if err == errCouldNotFindMCPSet {
-		ctrl.queue.AddAfter(key, 1*time.Minute)
-		return
-	}
-
 	if ctrl.queue.NumRequeues(key) < maxRetries {
 		glog.V(2).Infof("Error syncing kubeletconfig %v: %v", key, err)
 		ctrl.queue.AddRateLimited(key)
@@ -355,7 +344,16 @@ func (ctrl *Controller) syncStatusOnly(cfg *mcfgv1.KubeletConfig, err error, arg
 		if getErr != nil {
 			return getErr
 		}
-		newcfg.Status.Conditions = append(newcfg.Status.Conditions, wrapErrorWithCondition(err, args...))
+		// To avoid a long list of same statuses, only append a status if it is the first status
+		// or if the status message is different from the message of the last status recorded
+		// If the last status message is the same as the new one, then update the last status to
+		// reflect the latest time stamp from the new status message.
+		newStatusCondition := wrapErrorWithCondition(err, args...)
+		if len(newcfg.Status.Conditions) == 0 || newStatusCondition.Message != newcfg.Status.Conditions[len(newcfg.Status.Conditions)-1].Message {
+			newcfg.Status.Conditions = append(newcfg.Status.Conditions, newStatusCondition)
+		} else if newcfg.Status.Conditions[len(newcfg.Status.Conditions)-1].Message == newStatusCondition.Message {
+			newcfg.Status.Conditions[len(newcfg.Status.Conditions)-1] = newStatusCondition
+		}
 		_, lerr := ctrl.client.MachineconfigurationV1().KubeletConfigs().UpdateStatus(context.TODO(), newcfg, metav1.UpdateOptions{})
 		return lerr
 	})
@@ -365,8 +363,28 @@ func (ctrl *Controller) syncStatusOnly(cfg *mcfgv1.KubeletConfig, err error, arg
 	return err
 }
 
+// addAnnotation adds the annotions for a kubeletconfig object with the given annotationKey and annotationVal
+func (ctrl *Controller) addAnnotation(cfg *mcfgv1.KubeletConfig, annotationKey, annotationVal string) error {
+	annotationUpdateErr := retry.RetryOnConflict(updateBackoff, func() error {
+		newcfg, getErr := ctrl.mckLister.Get(cfg.Name)
+		if getErr != nil {
+			return getErr
+		}
+		newcfg.SetAnnotations(map[string]string{
+			annotationKey: annotationVal,
+		})
+		_, updateErr := ctrl.client.MachineconfigurationV1().KubeletConfigs().Update(context.TODO(), newcfg, metav1.UpdateOptions{})
+		return updateErr
+	})
+	if annotationUpdateErr != nil {
+		glog.Warningf("error updating the kubelet config with annotation key %q and value %q: %v", annotationKey, annotationVal, annotationUpdateErr)
+	}
+	return annotationUpdateErr
+}
+
 // syncKubeletConfig will sync the kubeletconfig with the given key.
 // This function is not meant to be invoked concurrently with the same key.
+//nolint:gocyclo
 func (ctrl *Controller) syncKubeletConfig(key string) error {
 	startTime := time.Now()
 	glog.V(4).Infof("Started syncing kubeletconfig %q (%v)", key, startTime)
@@ -440,47 +458,58 @@ func (ctrl *Controller) syncKubeletConfig(key string) error {
 	for _, pool := range mcpPools {
 		role := pool.Name
 		// Get MachineConfig
-		managedKey, err := getManagedKubeletConfigKey(pool, ctrl.client)
+		managedKey, err := getManagedKubeletConfigKey(pool, ctrl.client, cfg)
 		if err != nil {
-			return err
+			return ctrl.syncStatusOnly(cfg, err, "could not get kubelet config key: %v", err)
 		}
 		mc, err := ctrl.client.MachineconfigurationV1().MachineConfigs().Get(context.TODO(), managedKey, metav1.GetOptions{})
 		if err != nil && !macherrors.IsNotFound(err) {
 			return ctrl.syncStatusOnly(cfg, err, "could not find MachineConfig: %v", managedKey)
 		}
 		isNotFound := macherrors.IsNotFound(err)
-		// Generate the original KubeletConfig
-		originalKubeletIgn, err := ctrl.generateOriginalKubeletConfig(role)
-		if err != nil {
-			return ctrl.syncStatusOnly(cfg, err, "could not generate the original Kubelet config: %v", err)
-		}
-		dataURL, err := dataurl.DecodeString(*originalKubeletIgn.Contents.Source)
-		if err != nil {
-			return ctrl.syncStatusOnly(cfg, err, "could not decode the original Kubelet source string: %v", err)
-		}
-		originalKubeConfig, err := decodeKubeletConfig(dataURL.Data)
-		if err != nil {
-			return ctrl.syncStatusOnly(cfg, err, "could not deserialize the Kubelet source: %v", err)
-		}
-		specKubeletConfig, err := decodeKubeletConfig(cfg.Spec.KubeletConfig.Raw)
-		if err != nil {
-			return ctrl.syncStatusOnly(cfg, err, "could not deserialize the new Kubelet config: %v", err)
-		}
-		// Merge the Old and New
-		err = mergo.Merge(originalKubeConfig, specKubeletConfig, mergo.WithOverride)
-		if err != nil {
-			return ctrl.syncStatusOnly(cfg, err, "could not merge original config and new config: %v", err)
-		}
-		// Merge in Feature Gates
-		err = mergo.Merge(&originalKubeConfig.FeatureGates, featureGates, mergo.WithOverride)
-		if err != nil {
-			return ctrl.syncStatusOnly(cfg, err, "could not merge FeatureGates: %v", err)
-		}
-		// Encode the new config into raw JSON
-		cfgJSON, err := encodeKubeletConfig(originalKubeConfig, kubeletconfigv1beta1.SchemeGroupVersion)
-		if err != nil {
-			return ctrl.syncStatusOnly(cfg, err, "could not encode JSON: %v", err)
+
+		var kubeletIgnition *ign3types.File
+		var logLevelIgnition *ign3types.File
+
+		if cfg.Spec.KubeletConfig != nil && cfg.Spec.KubeletConfig.Raw != nil {
+			// Generate the original KubeletConfig
+			originalKubeletIgn, err := ctrl.generateOriginalKubeletConfig(role)
+			if err != nil {
+				return ctrl.syncStatusOnly(cfg, err, "could not generate the original Kubelet config: %v", err)
+			}
+			if originalKubeletIgn.Contents.Source == nil {
+				return ctrl.syncStatusOnly(cfg, err, "the original Kubelet source string is empty: %v", err)
+			}
+			dataURL, err := dataurl.DecodeString(*originalKubeletIgn.Contents.Source)
+			if err != nil {
+				return ctrl.syncStatusOnly(cfg, err, "could not decode the original Kubelet source string: %v", err)
+			}
+			originalKubeConfig, err := decodeKubeletConfig(dataURL.Data)
+			if err != nil {
+				return ctrl.syncStatusOnly(cfg, err, "could not deserialize the Kubelet source: %v", err)
+			}
+			specKubeletConfig, err := decodeKubeletConfig(cfg.Spec.KubeletConfig.Raw)
+			if err != nil {
+				return ctrl.syncStatusOnly(cfg, err, "could not deserialize the new Kubelet config: %v", err)
+			}
+			// Merge the Old and New
+			err = mergo.Merge(originalKubeConfig, specKubeletConfig, mergo.WithOverride)
+			if err != nil {
+				return ctrl.syncStatusOnly(cfg, err, "could not merge original config and new config: %v", err)
+			}
+			// Merge in Feature Gates
+			err = mergo.Merge(&originalKubeConfig.FeatureGates, featureGates, mergo.WithOverride)
+			if err != nil {
+				return ctrl.syncStatusOnly(cfg, err, "could not merge FeatureGates: %v", err)
+			}
+			// Encode the new config into raw JSON
+			cfgJSON, err := EncodeKubeletConfig(originalKubeConfig, kubeletconfigv1beta1.SchemeGroupVersion)
+			if err != nil {
+				return ctrl.syncStatusOnly(cfg, err, "could not encode JSON: %v", err)
+			}
+			kubeletIgnition = createNewKubeletIgnition(cfgJSON)
 		}
+
 		if isNotFound {
 			ignConfig := ctrlcommon.NewIgnConfig()
 			mc, err = ctrlcommon.MachineConfigFromIgnConfig(role, managedKey, ignConfig)
@@ -488,9 +517,30 @@ func (ctrl *Controller) syncKubeletConfig(key string) error {
 				return ctrl.syncStatusOnly(cfg, err, "could not create MachineConfig from new Ignition config: %v", err)
 			}
 			mc.ObjectMeta.UID = uuid.NewUUID()
+			_, ok := cfg.GetAnnotations()[ctrlcommon.MCNameSuffixAnnotationKey]
+			arr := strings.Split(managedKey, "-")
+			// If the MC name suffix annotation does not exist and the managed key value returned has a suffix, then add the MC name
+			// suffix annotation and suffix value to the kubelet config object
+			if len(arr) > 4 && !ok {
+				if err := ctrl.addAnnotation(cfg, ctrlcommon.MCNameSuffixAnnotationKey, arr[len(arr)-1]); err != nil {
+					return ctrl.syncStatusOnly(cfg, err, "could not update annotation for kubeletConfig")
+				}
+			}
 		}
-		cfgIgn := createNewKubeletIgnition(cfgJSON)
-		rawIgn, err := json.Marshal(cfgIgn)
+
+		if cfg.Spec.LogLevel != nil {
+			logLevelIgnition = createNewKubeletLogLevelIgnition(*cfg.Spec.LogLevel)
+		}
+
+		tempIgnConfig := ctrlcommon.NewIgnConfig()
+		if logLevelIgnition != nil {
+			tempIgnConfig.Storage.Files = append(tempIgnConfig.Storage.Files, *logLevelIgnition)
+		}
+		if kubeletIgnition != nil {
+			tempIgnConfig.Storage.Files = append(tempIgnConfig.Storage.Files, *kubeletIgnition)
+		}
+
+		rawIgn, err := json.Marshal(tempIgnConfig)
 		if err != nil {
 			return ctrl.syncStatusOnly(cfg, err, "could not marshal kubelet config Ignition: %v", err)
 		}
@@ -551,7 +601,7 @@ func (ctrl *Controller) popFinalizerFromKubeletConfig(kc *mcfgv1.KubeletConfig)
 		if err != nil {
 			return err
 		}
-		return ctrl.patchKubeletConfigsFunc(newcfg.Name, patch)
+		return ctrl.patchKubeletConfigs(newcfg.Name, patch)
 	})
 }
 
@@ -602,7 +652,7 @@ func (ctrl *Controller) addFinalizerToKubeletConfig(kc *mcfgv1.KubeletConfig, mc
 		if err != nil {
 			return err
 		}
-		return ctrl.patchKubeletConfigsFunc(newcfg.Name, patch)
+		return ctrl.patchKubeletConfigs(newcfg.Name, patch)
 	})
 }
 
