@@ -14,6 +14,7 @@ import (
 	"github.com/openshift/library-go/pkg/operator/v1helpers"
 	"github.com/openshift/machine-config-operator/internal"
 	mcfgv1 "github.com/openshift/machine-config-operator/pkg/apis/machineconfiguration.openshift.io/v1"
+	ctrlcommon "github.com/openshift/machine-config-operator/pkg/controller/common"
 	daemonconsts "github.com/openshift/machine-config-operator/pkg/daemon/constants"
 	mcfgclientset "github.com/openshift/machine-config-operator/pkg/generated/clientset/versioned"
 	"github.com/openshift/machine-config-operator/pkg/generated/clientset/versioned/scheme"
@@ -40,6 +41,9 @@ import (
 )
 
 const (
+	// WorkerLabel defines the label associated with worker node.
+	WorkerLabel = "node-role.kubernetes.io/worker"
+
 	// maxRetries is the number of times a machineconfig pool will be retried before it is dropped out of the queue.
 	// With the current rate-limiter in use (5ms*2^(maxRetries-1)) the following numbers represent the times
 	// a machineconfig pool is going to be requeued:
@@ -51,14 +55,14 @@ const (
 	// https://github.com/openshift/machine-config-operator/issues/301
 	updateDelay = 5 * time.Second
 
-	// masterLabel defines the label associated with master node. The master taint uses the same label as taint's key
-	masterLabel = "node-role.kubernetes.io/master"
-
-	// workerLabel defines the label associated with worker node.
-	workerLabel = "node-role.kubernetes.io/worker"
+	// osLabel is used to identify which type of OS the node has
+	osLabel = "kubernetes.io/os"
 
 	// schedulerCRName that we're interested in watching.
 	schedulerCRName = "cluster"
+
+	// masterPoolName is the control plane MachineConfigPool name
+	masterPoolName = "master"
 )
 
 var nodeUpdateBackoff = wait.Backoff{
@@ -156,7 +160,7 @@ func (ctrl *Controller) Run(workers int, stopCh <-chan struct{}) {
 }
 
 func (ctrl *Controller) getCurrentMasters() ([]*corev1.Node, error) {
-	nodeList, err := ctrl.nodeLister.List(labels.SelectorFromSet(labels.Set{masterLabel: ""}))
+	nodeList, err := ctrl.nodeLister.List(labels.SelectorFromSet(labels.Set{ctrlcommon.MasterLabel: ""}))
 	if err != nil {
 		return nil, fmt.Errorf("error while listing master nodes %v", err)
 	}
@@ -215,9 +219,6 @@ func (ctrl *Controller) checkMasterNodesOnUpdate(old, cur interface{}) {
 func (ctrl *Controller) makeMastersUnSchedulable(currentMasters []*corev1.Node) []error {
 	var errs []error
 	for _, node := range currentMasters {
-		if !CheckMasterIsAlreadySchedulable(node) {
-			continue
-		}
 		if err := ctrl.makeMasterNodeUnSchedulable(node); err != nil {
 			errs = append(errs, fmt.Errorf("failed making node %v schedulable with error %v", node.Name, err))
 		}
@@ -232,15 +233,23 @@ func (ctrl *Controller) makeMasterNodeUnSchedulable(node *corev1.Node) error {
 	_, err := internal.UpdateNodeRetry(ctrl.kubeClient.CoreV1().Nodes(), ctrl.nodeLister, node.Name, func(node *corev1.Node) {
 		// Remove worker label
 		newLabels := node.Labels
-		if _, hasWorkerLabel := newLabels[workerLabel]; hasWorkerLabel {
-			delete(newLabels, workerLabel)
+		if _, hasWorkerLabel := newLabels[WorkerLabel]; hasWorkerLabel {
+			delete(newLabels, WorkerLabel)
 		}
 		node.Labels = newLabels
 		// Add master taint
-		newTaints := node.Spec.Taints
-		masterUnSchedulableTaint := corev1.Taint{Key: masterLabel, Effect: corev1.TaintEffectNoSchedule}
-		newTaints = append(newTaints, masterUnSchedulableTaint)
-		node.Spec.Taints = newTaints
+		hasMasterTaint := false
+		for _, taint := range node.Spec.Taints {
+			if taint.Key == ctrlcommon.MasterLabel && taint.Effect == corev1.TaintEffectNoSchedule {
+				hasMasterTaint = true
+			}
+		}
+		if !hasMasterTaint {
+			newTaints := node.Spec.Taints
+			masterUnSchedulableTaint := corev1.Taint{Key: ctrlcommon.MasterLabel, Effect: corev1.TaintEffectNoSchedule}
+			newTaints = append(newTaints, masterUnSchedulableTaint)
+			node.Spec.Taints = newTaints
+		}
 	})
 	if err != nil {
 		return err
@@ -248,33 +257,20 @@ func (ctrl *Controller) makeMasterNodeUnSchedulable(node *corev1.Node) error {
 	return nil
 }
 
-// CheckMasterIsAlreadySchedulable checks if the given node has a worker label and doesn't have NoSchedule master
-// taint
-func CheckMasterIsAlreadySchedulable(node *corev1.Node) bool {
-	_, hasWorkerLabel := node.Labels[workerLabel]
-	hasMasterTaint := false
-	for _, taint := range node.Spec.Taints {
-		if taint.Key == masterLabel && taint.Effect == corev1.TaintEffectNoSchedule {
-			hasMasterTaint = true
-		}
-	}
-	return hasWorkerLabel && !hasMasterTaint
-}
-
 // makeMasterNodeSchedulable makes master node schedulable by removing NoSchedule master taint and
 // adding worker label
 func (ctrl *Controller) makeMasterNodeSchedulable(node *corev1.Node) error {
 	_, err := internal.UpdateNodeRetry(ctrl.kubeClient.CoreV1().Nodes(), ctrl.nodeLister, node.Name, func(node *corev1.Node) {
 		// Add worker label
 		newLabels := node.Labels
-		if _, hasWorkerLabel := newLabels[workerLabel]; !hasWorkerLabel {
-			newLabels[workerLabel] = ""
+		if _, hasWorkerLabel := newLabels[WorkerLabel]; !hasWorkerLabel {
+			newLabels[WorkerLabel] = ""
 		}
 		node.Labels = newLabels
 		// Remove master taint
 		newTaints := []corev1.Taint{}
 		for _, t := range node.Spec.Taints {
-			if t.Key == masterLabel && t.Effect == corev1.TaintEffectNoSchedule {
+			if t.Key == ctrlcommon.MasterLabel && t.Effect == corev1.TaintEffectNoSchedule {
 				continue
 			}
 			newTaints = append(newTaints, t)
@@ -335,10 +331,24 @@ func (ctrl *Controller) getMastersSchedulable() (bool, error) {
 
 // Determine if a given Node is a master
 func (ctrl *Controller) isMaster(node *corev1.Node) bool {
-	_, master := node.ObjectMeta.Labels[masterLabel]
+	_, master := node.ObjectMeta.Labels[ctrlcommon.MasterLabel]
 	return master
 }
 
+// isWindows checks if given node is a Windows node or a Linux node
+func isWindows(node *corev1.Node) bool {
+	windowsOsValue := "windows"
+	if value, ok := node.ObjectMeta.Labels[osLabel]; ok {
+		if value == windowsOsValue {
+			return true
+		}
+		return false
+	}
+	// All the nodes should have a OS label populated by kubelet, if not just to maintain
+	// backwards compatibility, we can returning true here.
+	return false
+}
+
 // Given a master Node, ensure it reflects the current mastersSchedulable setting
 func (ctrl *Controller) reconcileMaster(node *corev1.Node) {
 	mastersSchedulable, err := ctrl.getMastersSchedulable()
@@ -346,14 +356,13 @@ func (ctrl *Controller) reconcileMaster(node *corev1.Node) {
 		goerrs.Wrap(err, "Getting scheduler config failed")
 		return
 	}
-	alreadySchedulable := CheckMasterIsAlreadySchedulable(node)
-	if mastersSchedulable && !alreadySchedulable {
+	if mastersSchedulable {
 		err = ctrl.makeMasterNodeSchedulable(node)
 		if err != nil {
 			goerrs.Wrap(err, "Failed making master Node schedulable")
 			return
 		}
-	} else if !mastersSchedulable && alreadySchedulable {
+	} else if !mastersSchedulable {
 		err = ctrl.makeMasterNodeUnSchedulable(node)
 		if err != nil {
 			goerrs.Wrap(err, "Failed making master Node unschedulable")
@@ -363,6 +372,7 @@ func (ctrl *Controller) reconcileMaster(node *corev1.Node) {
 }
 
 // Get a list of current masters and apply scheduler config to them
+// TODO: Taint reconciliation should happen elsewhere, in a generic taint/label reconciler
 func (ctrl *Controller) reconcileMasters() {
 	currentMasters, err := ctrl.getCurrentMasters()
 	if err != nil {
@@ -399,6 +409,16 @@ func (ctrl *Controller) addNode(obj interface{}) {
 	}
 }
 
+func (ctrl *Controller) logPool(pool *mcfgv1.MachineConfigPool, format string, args ...interface{}) {
+	msg := fmt.Sprintf(format, args...)
+	glog.Infof("Pool %s: %s", pool.Name, msg)
+}
+
+func (ctrl *Controller) logPoolNode(pool *mcfgv1.MachineConfigPool, node *corev1.Node, format string, args ...interface{}) {
+	msg := fmt.Sprintf(format, args...)
+	glog.Infof("Pool %s: node %s: %s", pool.Name, node.Name, msg)
+}
+
 func (ctrl *Controller) updateNode(old, cur interface{}) {
 	oldNode := old.(*corev1.Node)
 	curNode := cur.(*corev1.Node)
@@ -431,16 +451,16 @@ func (ctrl *Controller) updateNode(old, cur interface{}) {
 	if oldReady != newReady {
 		changed = true
 		if newReadyErr != nil {
-			glog.Infof("Pool %s: node %s is now reporting unready: %v", pool.Name, curNode.Name, newReadyErr)
+			ctrl.logPoolNode(pool, curNode, "Reporting unready: %v", newReadyErr)
 		} else {
-			glog.Infof("Pool %s: node %s is now reporting ready", pool.Name, curNode.Name)
+			ctrl.logPoolNode(pool, curNode, "Reporting ready")
 		}
 	}
 
 	// Specifically log when a node has completed an update so the MCC logs are a useful central aggregate of state changes
 	if oldNode.Annotations[daemonconsts.CurrentMachineConfigAnnotationKey] != oldNode.Annotations[daemonconsts.DesiredMachineConfigAnnotationKey] &&
 		isNodeDone(curNode) {
-		glog.Infof("Pool %s: node %s has completed update to %s", pool.Name, curNode.Name, curNode.Annotations[daemonconsts.DesiredMachineConfigAnnotationKey])
+		ctrl.logPoolNode(pool, curNode, "Completed update to %s", curNode.Annotations[daemonconsts.DesiredMachineConfigAnnotationKey])
 		changed = true
 	} else {
 		annos := []string{
@@ -449,13 +469,18 @@ func (ctrl *Controller) updateNode(old, cur interface{}) {
 			daemonconsts.MachineConfigDaemonStateAnnotationKey,
 		}
 		for _, anno := range annos {
-			if oldNode.Annotations[anno] != curNode.Annotations[anno] {
-				glog.Infof("Pool %s: node %s changed %s = %s", pool.Name, curNode.Name, anno, curNode.Annotations[anno])
+			newValue := curNode.Annotations[anno]
+			if oldNode.Annotations[anno] != newValue {
+				ctrl.logPoolNode(pool, curNode, "changed annotation %s = %s", anno, newValue)
 				changed = true
+				// For the control plane, emit events for these since they're important
+				if pool.Name == masterPoolName {
+					ctrl.eventRecorder.Eventf(pool, corev1.EventTypeNormal, "AnnotationChange", "Node %s now has %s=%s", curNode.Name, anno, newValue)
+				}
 			}
 		}
 		if !reflect.DeepEqual(oldNode.Labels, curNode.Labels) {
-			glog.Infof("Pool %s: node %s changed labels", pool.Name, curNode.Name)
+			ctrl.logPoolNode(pool, curNode, "changed labels")
 			changed = true
 		}
 	}
@@ -510,7 +535,14 @@ func (ctrl *Controller) deleteNode(obj interface{}) {
 // getPoolsForNode chooses the MachineConfigPools that should be used for a given node.
 // It disambiguates in the case where e.g. a node has both master/worker roles applied,
 // and where a custom role may be used. It returns a slice of all the pools the node belongs to.
+// It also ignores the Windows nodes.
 func (ctrl *Controller) getPoolsForNode(node *corev1.Node) ([]*mcfgv1.MachineConfigPool, error) {
+	if isWindows(node) {
+		// This is not an error, is this a Windows Node and it won't be managed by MCO. We're explicitly logging
+		// here at a high level to disambiguate this from other pools = nil  scenario
+		glog.V(4).Infof("Node %v is a windows node so won't be managed by MCO", node.Name)
+		return nil, nil
+	}
 	pl, err := ctrl.mcpLister.List(labels.Everything())
 	if err != nil {
 		return nil, err
@@ -539,7 +571,7 @@ func (ctrl *Controller) getPoolsForNode(node *corev1.Node) ([]*mcfgv1.MachineCon
 	var master, worker *mcfgv1.MachineConfigPool
 	var custom []*mcfgv1.MachineConfigPool
 	for _, pool := range pools {
-		if pool.Name == "master" {
+		if pool.Name == masterPoolName {
 			master = pool
 		} else if pool.Name == "worker" {
 			worker = pool
@@ -704,22 +736,35 @@ func (ctrl *Controller) syncMachineConfigPool(key string) error {
 	}
 
 	if pool.Spec.Paused {
+		if mcfgv1.IsMachineConfigPoolConditionTrue(pool.Status.Conditions, mcfgv1.MachineConfigPoolUpdating) {
+			glog.Infof("Pool %s is paused and will not update.", pool.Name)
+		}
 		return ctrl.syncStatusOnly(pool)
 	}
 
 	nodes, err := ctrl.getNodesForPool(pool)
 	if err != nil {
+		if syncErr := ctrl.syncStatusOnly(pool); syncErr != nil {
+			return goerrs.Wrapf(err, "error getting nodes for pool %q, sync error: %v", pool.Name, syncErr)
+		}
 		return err
 	}
 
 	maxunavail, err := maxUnavailable(pool, nodes)
 	if err != nil {
+		if syncErr := ctrl.syncStatusOnly(pool); syncErr != nil {
+			return goerrs.Wrapf(err, "error getting max unavailable count for pool %q, sync error: %v", pool.Name, syncErr)
+		}
 		return err
 	}
 
-	candidates := getCandidateMachines(pool, nodes, maxunavail)
-	for _, node := range candidates {
-		if err := ctrl.setDesiredMachineConfigAnnotation(node.Name, pool.Spec.Configuration.Name); err != nil {
+	candidates, capacity := getAllCandidateMachines(pool, nodes, maxunavail)
+	if len(candidates) > 0 {
+		ctrl.logPool(pool, "%d candidate nodes for update, capacity: %d", len(candidates), capacity)
+		if err := ctrl.updateCandidateMachines(pool, candidates, capacity); err != nil {
+			if syncErr := ctrl.syncStatusOnly(pool); syncErr != nil {
+				return goerrs.Wrapf(err, "error setting desired machine config annotation for pool %q, sync error: %v", pool.Name, syncErr)
+			}
 			return err
 		}
 	}
@@ -756,7 +801,6 @@ func (ctrl *Controller) getNodesForPool(pool *mcfgv1.MachineConfigPool) ([]*core
 }
 
 func (ctrl *Controller) setDesiredMachineConfigAnnotation(nodeName, currentConfig string) error {
-	glog.Infof("Setting node %s to desired config %s", nodeName, currentConfig)
 	return clientretry.RetryOnConflict(nodeUpdateBackoff, func() error {
 		oldNode, err := ctrl.kubeClient.CoreV1().Nodes().Get(context.TODO(), nodeName, metav1.GetOptions{})
 		if err != nil {
@@ -789,13 +833,15 @@ func (ctrl *Controller) setDesiredMachineConfigAnnotation(nodeName, currentConfi
 	})
 }
 
-func getCandidateMachines(pool *mcfgv1.MachineConfigPool, nodesInPool []*corev1.Node, maxUnavailable int) []*corev1.Node {
+// getAllCandidateMachines returns all possible nodes which can be updated to the target config, along with a maximum
+// capacity.  It is the reponsibility of the caller to choose a subset of the nodes given the capacity.
+func getAllCandidateMachines(pool *mcfgv1.MachineConfigPool, nodesInPool []*corev1.Node, maxUnavailable int) ([]*corev1.Node, uint) {
 	targetConfig := pool.Spec.Configuration.Name
 
 	unavail := getUnavailableMachines(nodesInPool)
 	// If we're at capacity, there's nothing to do.
 	if len(unavail) >= maxUnavailable {
-		return nil
+		return nil, 0
 	}
 	capacity := maxUnavailable - len(unavail)
 	failingThisConfig := 0
@@ -816,16 +862,83 @@ func getCandidateMachines(pool *mcfgv1.MachineConfigPool, nodesInPool []*corev1.
 	// availability - it might be a transient issue, and if the issue
 	// clears we don't want multiple to update at once.
 	if failingThisConfig >= capacity {
-		return nil
+		return nil, 0
 	}
 	capacity -= failingThisConfig
 
-	if len(nodes) < capacity {
+	return nodes, uint(capacity)
+}
+
+// getCandidateMachines returns the maximum subset of nodes which can be updated to the target config given availability constraints.
+func getCandidateMachines(pool *mcfgv1.MachineConfigPool, nodesInPool []*corev1.Node, maxUnavailable int) []*corev1.Node {
+	nodes, capacity := getAllCandidateMachines(pool, nodesInPool, maxUnavailable)
+	if uint(len(nodes)) < capacity {
 		return nodes
 	}
 	return nodes[:capacity]
 }
 
+// getCurrentEtcdLeader is not yet implemented
+func (ctrl *Controller) getCurrentEtcdLeader(candidates []*corev1.Node) (*corev1.Node, error) {
+	return nil, nil
+}
+
+// filterControlPlaneCandidateNodes adjusts the candidates and capacity specifically
+// for the control plane, e.g. based on which node is the etcd leader at the time.
+// nolint:unparam
+func (ctrl *Controller) filterControlPlaneCandidateNodes(pool *mcfgv1.MachineConfigPool, candidates []*corev1.Node, capacity uint) ([]*corev1.Node, uint, error) {
+	if len(candidates) <= 1 {
+		return candidates, capacity, nil
+	}
+	etcdLeader, err := ctrl.getCurrentEtcdLeader(candidates)
+	if err != nil {
+		glog.Warningf("Failed to find current etcd leader (continuing anyways): %v", err)
+	}
+	var newCandidates []*corev1.Node
+	for _, node := range candidates {
+		if node == etcdLeader {
+			// For now make this an event so we know it's working, even though it's more of a non-event
+			ctrl.eventRecorder.Eventf(pool, corev1.EventTypeNormal, "DeferringEtcdLeaderUpdate", "Deferring update of etcd leader %s", node.Name)
+			glog.Infof("Deferring update of etcd leader: %s", node.Name)
+			continue
+		}
+		newCandidates = append(newCandidates, node)
+	}
+	return newCandidates, capacity, nil
+}
+
+// updateCandidateMachines sets the desiredConfig annotation the candidate machines
+func (ctrl *Controller) updateCandidateMachines(pool *mcfgv1.MachineConfigPool, candidates []*corev1.Node, capacity uint) error {
+	if pool.Name == masterPoolName {
+		var err error
+		candidates, capacity, err = ctrl.filterControlPlaneCandidateNodes(pool, candidates, capacity)
+		if err != nil {
+			return err
+		}
+		// In practice right now these counts will be 1 but let's stay general to support 5 etcd nodes in the future
+		ctrl.logPool(pool, "filtered to %d candidate nodes for update, capacity: %d", len(candidates), capacity)
+	}
+	if capacity < uint(len(candidates)) {
+		// Arbitrarily pick the first N candidates; no attempt at sorting.
+		// Perhaps later we allow admins to weight somehow, or do something more intelligent.
+		candidates = candidates[:capacity]
+	}
+	targetConfig := pool.Spec.Configuration.Name
+	for _, node := range candidates {
+		ctrl.logPool(pool, "Setting node %s target to %s", node.Name, targetConfig)
+		if err := ctrl.setDesiredMachineConfigAnnotation(node.Name, targetConfig); err != nil {
+			return goerrs.Wrapf(err, "setting desired config for node %s", node.Name)
+		}
+	}
+	if len(candidates) == 1 {
+		candidate := candidates[0]
+		ctrl.eventRecorder.Eventf(pool, corev1.EventTypeNormal, "SetDesiredConfig", "Targeted node %s to config %s", candidate.Name, targetConfig)
+	} else {
+		ctrl.eventRecorder.Eventf(pool, corev1.EventTypeNormal, "SetDesiredConfig", "Set target for %d nodes to config %s", targetConfig)
+	}
+	return nil
+}
+
 func maxUnavailable(pool *mcfgv1.MachineConfigPool, nodes []*corev1.Node) (int, error) {
 	intOrPercent := intstrutil.FromInt(1)
 	if pool.Spec.MaxUnavailable != nil {
@@ -838,7 +951,7 @@ func maxUnavailable(pool *mcfgv1.MachineConfigPool, nodes []*corev1.Node) (int,
 	if maxunavail == 0 {
 		maxunavail = 1
 	}
-	if pool.Name == "master" {
+	if pool.Name == masterPoolName {
 		// calculate the fault tolerance dynamically for the master pool
 		// to avoid risking losing etcd quorum.
 		tolerance := len(nodes) - ((len(nodes) / 2) + 1)
