@@ -16,7 +16,7 @@ import (
 	"time"
 
 	"github.com/clarketm/json"
-	ign3types "github.com/coreos/ignition/v2/config/v3_1/types"
+	ign3types "github.com/coreos/ignition/v2/config/v3_2/types"
 	"github.com/golang/glog"
 	"github.com/google/renameio"
 	errors "github.com/pkg/errors"
@@ -48,6 +48,16 @@ const (
 	fipsFile              = "/proc/sys/crypto/fips_enabled"
 	extensionsRepo        = "/etc/yum.repos.d/coreos-extensions.repo"
 	osImageContentBaseDir = "/run/mco-machine-os-content/"
+
+	// These are the actions for a node to take after applying config changes. (e.g. a new machineconfig is applied)
+	// "None" means no special action needs to be taken. A drain will still happen.
+	// This currently happens when ssh keys or pull secret (/var/lib/kubelet/config.json) is changed
+	postConfigChangeActionNone = "none"
+	// Rebooting is still the default scenario for any other change
+	postConfigChangeActionReboot = "reboot"
+	// Crio reload will happen when /etc/containers/registries.conf is changed. This will cause
+	// a "systemctl reload crio"
+	postConfigChangeActionReloadCrio = "reload crio"
 )
 
 func writeFileAtomicallyWithDefaults(fpath string, b []byte) error {
@@ -93,9 +103,70 @@ func getNodeRef(node *corev1.Node) *corev1.ObjectReference {
 	}
 }
 
-// finalizeAndReboot is the last step in an update(), and it can also
-// be called as a special case for the "bootstrap pivot".
-func (dn *Daemon) finalizeAndReboot(newConfig *mcfgv1.MachineConfig) (retErr error) {
+func reloadService(name string) error {
+	_, err := runGetOut("systemctl", "reload", name)
+	return err
+}
+
+// performPostConfigChangeAction takes action based on what postConfigChangeAction has been asked.
+// For non-reboot action, it applies configuration, updates node's config and state.
+// In the end uncordon node to schedule workload.
+// If at any point an error occurs, we reboot the node so that node has correct configuration.
+func (dn *Daemon) performPostConfigChangeAction(postConfigChangeActions []string, configName string) error {
+	if ctrlcommon.InSlice(postConfigChangeActionReboot, postConfigChangeActions) {
+		dn.logSystem("Rebooting node")
+		return dn.reboot(fmt.Sprintf("Node will reboot into config %s", configName))
+	}
+
+	if ctrlcommon.InSlice(postConfigChangeActionNone, postConfigChangeActions) {
+		if dn.recorder != nil {
+			dn.recorder.Eventf(getNodeRef(dn.node), corev1.EventTypeNormal, "SkipReboot", "Config changes do not require reboot.")
+		}
+		dn.logSystem("Node has Desired Config %s, skipping reboot", configName)
+	}
+
+	if ctrlcommon.InSlice(postConfigChangeActionReloadCrio, postConfigChangeActions) {
+		serviceName := "crio"
+
+		if err := reloadService(serviceName); err != nil {
+			if dn.recorder != nil {
+				dn.recorder.Eventf(getNodeRef(dn.node), corev1.EventTypeWarning, "FailedServiceReload", fmt.Sprintf("Reloading %s service failed. Error: %v", serviceName, err))
+			}
+			dn.logSystem("Reloading %s configuration failed, node will reboot instead. Error: %v", serviceName, err)
+			dn.reboot(fmt.Sprintf("Node will reboot into config %s", configName))
+		}
+
+		if dn.recorder != nil {
+			dn.recorder.Eventf(getNodeRef(dn.node), corev1.EventTypeNormal, "SkipReboot", "Config changes do not require reboot. Service %s was reloaded.", serviceName)
+		}
+		dn.logSystem("%s config reloaded successfully! Desired config %s has been applied, skipping reboot", serviceName, configName)
+	}
+
+	// We are here, which means reboot was not needed to apply the configuration.
+
+	// Get current state of node, in case of an error reboot
+	state, err := dn.getStateAndConfigs(configName)
+	if err != nil {
+		glog.Errorf("Error processing state and configs, node will reboot instead. Error: %v", err)
+		return dn.reboot(fmt.Sprintf("Node will reboot into config %s", configName))
+	}
+
+	var inDesiredConfig bool
+	if inDesiredConfig, err = dn.updateConfigAndState(state); err != nil {
+		glog.Errorf("Setting node's state to Done failed, node will reboot instead. Error: %v", err)
+		return dn.reboot(fmt.Sprintf("Node will reboot into config %s", configName))
+	}
+	if inDesiredConfig {
+		return nil
+	}
+
+	// currentConfig != desiredConfig, kick off an update
+	return dn.triggerUpdateWithMachineConfig(state.currentConfig, state.desiredConfig)
+}
+
+// finalizeBeforeReboot is the last step in an update() and then we take appropriate postConfigChangeAction.
+// It can also be called as a special case for the "bootstrap pivot".
+func (dn *Daemon) finalizeBeforeReboot(newConfig *mcfgv1.MachineConfig) (retErr error) {
 	if out, err := dn.storePendingState(newConfig, 1); err != nil {
 		return errors.Wrapf(err, "failed to log pending config: %s", string(out))
 	}
@@ -114,8 +185,7 @@ func (dn *Daemon) finalizeAndReboot(newConfig *mcfgv1.MachineConfig) (retErr err
 		dn.recorder.Eventf(getNodeRef(dn.node), corev1.EventTypeNormal, "PendingConfig", fmt.Sprintf("Written pending config %s", newConfig.GetName()))
 	}
 
-	// reboot. this function shouldn't actually return.
-	return dn.reboot(fmt.Sprintf("Node will reboot into config %v", newConfig.GetName()))
+	return nil
 }
 
 func (dn *Daemon) drain() error {
@@ -351,7 +421,7 @@ func (dn *Daemon) applyOSChanges(oldConfig, newConfig *mcfgv1.MachineConfig) (re
 		// Delete extracted OS image once we are done.
 		defer os.RemoveAll(osImageContentDir)
 
-		if dn.OperatingSystem == MachineConfigDaemonOSRHCOS || dn.OperatingSystem == MachineConfigDaemonOSFCOS {
+		if dn.os.IsCoreOSVariant() {
 			if err := addExtensionsRepo(osImageContentDir); err != nil {
 				return err
 			}
@@ -381,8 +451,10 @@ func (dn *Daemon) applyOSChanges(oldConfig, newConfig *mcfgv1.MachineConfig) (re
 	}()
 
 	// Apply kargs
-	if err := dn.updateKernelArguments(oldConfig, newConfig); err != nil {
-		return err
+	if mcDiff.kargs {
+		if err := dn.updateKernelArguments(oldConfig, newConfig); err != nil {
+			return err
+		}
 	}
 
 	// Switch to real time kernel
@@ -402,6 +474,101 @@ func (dn *Daemon) applyOSChanges(oldConfig, newConfig *mcfgv1.MachineConfig) (re
 
 }
 
+func calculatePostConfigChangeActionFromFileDiffs(oldIgnConfig, newIgnConfig ign3types.Config) (actions []string) {
+	filesPostConfigChangeActionNone := []string{
+		"/var/lib/kubelet/config.json",
+	}
+	filesPostConfigChangeActionReloadCrio := []string{
+		"/etc/containers/registries.conf",
+	}
+
+	oldFileSet := make(map[string]ign3types.File)
+	for _, f := range oldIgnConfig.Storage.Files {
+		oldFileSet[f.Path] = f
+	}
+	newFileSet := make(map[string]ign3types.File)
+	for _, f := range newIgnConfig.Storage.Files {
+		newFileSet[f.Path] = f
+	}
+	diffFileSet := []string{}
+
+	// First check if any files were removed
+	for path := range oldFileSet {
+		_, ok := newFileSet[path]
+		if !ok {
+			// debug: remove
+			glog.Infof("File diff: %v was deleted", path)
+			diffFileSet = append(diffFileSet, path)
+		}
+	}
+
+	// Now check if any files were added/changed
+	for path, newFile := range newFileSet {
+		oldFile, ok := oldFileSet[path]
+		if !ok {
+			// debug: remove
+			glog.Infof("File diff: %v was added", path)
+			diffFileSet = append(diffFileSet, path)
+		} else if !reflect.DeepEqual(oldFile, newFile) {
+			// debug: remove
+			glog.Infof("File diff: detected change to %v", newFile.Path)
+			diffFileSet = append(diffFileSet, path)
+		}
+	}
+
+	// Now calculate action
+	for _, k := range diffFileSet {
+		if ctrlcommon.InSlice(k, filesPostConfigChangeActionNone) {
+			continue
+		} else if ctrlcommon.InSlice(k, filesPostConfigChangeActionReloadCrio) {
+			actions = []string{postConfigChangeActionReloadCrio}
+			continue
+		} else {
+			actions = []string{postConfigChangeActionReboot}
+			break
+		}
+	}
+
+	if len(actions) == 0 {
+		actions = []string{postConfigChangeActionNone}
+	}
+	return
+}
+
+func calculatePostConfigChangeAction(oldConfig, newConfig *mcfgv1.MachineConfig) ([]string, error) {
+	// If a machine-config-daemon-force file is present, it means the user wants to
+	// move to desired state without additional validation. We will reboot the node in
+	// this case regardless of what MachineConfig diff is.
+	if _, err := os.Stat(constants.MachineConfigDaemonForceFile); err == nil {
+		if err := os.Remove(constants.MachineConfigDaemonForceFile); err != nil {
+			return []string{}, errors.Wrap(err, "failed to remove force validation file")
+		}
+		glog.Infof("Setting post config change action to postConfigChangeActionReboot; %s present", constants.MachineConfigDaemonForceFile)
+		return []string{postConfigChangeActionReboot}, nil
+	}
+
+	diff, err := newMachineConfigDiff(oldConfig, newConfig)
+	if err != nil {
+		return []string{}, err
+	}
+	if diff.osUpdate || diff.kargs || diff.fips || diff.units || diff.kernelType || diff.extensions {
+		// must reboot
+		return []string{postConfigChangeActionReboot}, nil
+	}
+
+	oldIgnConfig, err := ctrlcommon.ParseAndConvertConfig(oldConfig.Spec.Config.Raw)
+	if err != nil {
+		return []string{}, err
+	}
+	newIgnConfig, err := ctrlcommon.ParseAndConvertConfig(newConfig.Spec.Config.Raw)
+	if err != nil {
+		return []string{}, err
+	}
+
+	// We don't actually have to consider ssh keys changes, which is the only section of passwd that is allowed to change
+	return calculatePostConfigChangeActionFromFileDiffs(oldIgnConfig, newIgnConfig), nil
+}
+
 // update the node to the provided node configuration.
 func (dn *Daemon) update(oldConfig, newConfig *mcfgv1.MachineConfig) (retErr error) {
 	oldConfig = canonicalizeEmptyMC(oldConfig)
@@ -448,6 +615,11 @@ func (dn *Daemon) update(oldConfig, newConfig *mcfgv1.MachineConfig) (retErr err
 
 	dn.logSystem("Starting update from %s to %s: %+v", oldConfigName, newConfigName, diff)
 
+	actions, err := calculatePostConfigChangeAction(oldConfig, newConfig)
+	if err != nil {
+		return err
+	}
+
 	if err := dn.drain(); err != nil {
 		return err
 	}
@@ -523,7 +695,11 @@ func (dn *Daemon) update(oldConfig, newConfig *mcfgv1.MachineConfig) (retErr err
 		glog.Info("Updated kernel tuning arguments")
 	}
 
-	return dn.finalizeAndReboot(newConfig)
+	if err := dn.finalizeBeforeReboot(newConfig); err != nil {
+		return err
+	}
+
+	return dn.performPostConfigChangeAction(actions, newConfig.GetName())
 }
 
 // machineConfigDiff represents an ad-hoc difference between two MachineConfig objects.
@@ -638,21 +814,20 @@ func reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) (*machineConfigDif
 			return nil, errors.New("ignition Passwd Groups section contains changes")
 		}
 		if !reflect.DeepEqual(oldIgn.Passwd.Users, newIgn.Passwd.Users) {
-			// check if the prior config is empty and that this is the first time running.
-			// if so, the SSHKey from the cluster config and user "core" must be added to machine config.
-			if len(oldIgn.Passwd.Users) > 0 && len(newIgn.Passwd.Users) >= 1 {
-				// there is an update to Users, we must verify that it is ONLY making an acceptable
-				// change to the SSHAuthorizedKeys for the user "core"
-				for _, user := range newIgn.Passwd.Users {
-					if user.Name != coreUserName {
-						return nil, errors.New("ignition passwd user section contains unsupported changes: non-core user")
-					}
-				}
-				glog.Infof("user data to be verified before ssh update: %v", newIgn.Passwd.Users[len(newIgn.Passwd.Users)-1])
-				if err := verifyUserFields(newIgn.Passwd.Users[len(newIgn.Passwd.Users)-1]); err != nil {
-					return nil, err
+			if len(oldIgn.Passwd.Users) > 0 && len(newIgn.Passwd.Users) == 0 {
+				return nil, errors.New("ignition passwd user section contains unsupported changes: user core may not be deleted")
+			}
+			// there is an update to Users, we must verify that it is ONLY making an acceptable
+			// change to the SSHAuthorizedKeys for the user "core"
+			for _, user := range newIgn.Passwd.Users {
+				if user.Name != coreUserName {
+					return nil, errors.New("ignition passwd user section contains unsupported changes: non-core user")
 				}
 			}
+			glog.Infof("user data to be verified before ssh update: %v", newIgn.Passwd.Users[len(newIgn.Passwd.Users)-1])
+			if err := verifyUserFields(newIgn.Passwd.Users[len(newIgn.Passwd.Users)-1]); err != nil {
+				return nil, err
+			}
 		}
 	}
 
@@ -816,12 +991,12 @@ func parseKernelArguments(kargs []string) []string {
 	return parsed
 }
 
-// generateKargsCommand performs a diff between the old/new MC kernelArguments,
+// generateKargs performs a diff between the old/new MC kernelArguments,
 // and generates the command line arguments suitable for `rpm-ostree kargs`.
 // Note what we really should be doing though is also looking at the *current*
 // kernel arguments in case there was drift.  But doing that requires us knowing
 // what the "base" arguments are. See https://github.com/ostreedev/ostree/issues/479
-func generateKargsCommand(oldConfig, newConfig *mcfgv1.MachineConfig) []string {
+func generateKargs(oldConfig, newConfig *mcfgv1.MachineConfig) []string {
 	oldKargs := parseKernelArguments(oldConfig.Spec.KernelArguments)
 	newKargs := parseKernelArguments(newConfig.Spec.KernelArguments)
 	cmdArgs := []string{}
@@ -841,17 +1016,18 @@ func generateKargsCommand(oldConfig, newConfig *mcfgv1.MachineConfig) []string {
 
 // updateKernelArguments adjusts the kernel args
 func (dn *Daemon) updateKernelArguments(oldConfig, newConfig *mcfgv1.MachineConfig) error {
-	diff := generateKargsCommand(oldConfig, newConfig)
-	if len(diff) == 0 {
+	kargs := generateKargs(oldConfig, newConfig)
+	if len(kargs) == 0 {
 		return nil
 	}
-	if dn.OperatingSystem != MachineConfigDaemonOSRHCOS && dn.OperatingSystem != MachineConfigDaemonOSFCOS {
-		return fmt.Errorf("updating kargs on non-CoreOS nodes is not supported: %v", diff)
+	if !dn.os.IsCoreOSVariant() {
+		return fmt.Errorf("updating kargs on non-CoreOS nodes is not supported: %v", kargs)
 	}
 
-	args := append([]string{"kargs"}, diff...)
+	args := append([]string{"kargs"}, kargs...)
 	dn.logSystem("Running rpm-ostree %v", args)
-	return exec.Command("rpm-ostree", args...).Run()
+	_, err := runGetOut("rpm-ostree", args...)
+	return err
 }
 
 func (dn *Daemon) generateExtensionsArgs(oldConfig, newConfig *mcfgv1.MachineConfig) []string {
@@ -883,7 +1059,7 @@ func (dn *Daemon) generateExtensionsArgs(oldConfig, newConfig *mcfgv1.MachineCon
 
 	extArgs := []string{"update"}
 
-	if dn.OperatingSystem == MachineConfigDaemonOSRHCOS {
+	if dn.os.IsRHCOS() {
 		extensions := getSupportedExtensions()
 		for _, ext := range added {
 			for _, pkg := range extensions[ext] {
@@ -902,7 +1078,7 @@ func (dn *Daemon) generateExtensionsArgs(oldConfig, newConfig *mcfgv1.MachineCon
 	// See https://github.com/openshift/release/blob/959c2954344438c4eed3ec7f52a5e099e8335516/ci-operator/jobs/openshift/release/openshift-release-release-4.7-periodics.yaml#L586
 	// TODO: Once the package list has been stabilized, we can make use of the group and add
 	// all the packages required to enable OKD as a single extension.
-	if dn.OperatingSystem == MachineConfigDaemonOSFCOS {
+	if dn.os.IsFCOS() {
 		for _, ext := range added {
 			extArgs = append(extArgs, "--install", ext)
 		}
@@ -949,23 +1125,20 @@ func (dn *Daemon) applyExtensions(oldConfig, newConfig *mcfgv1.MachineConfig) er
 		return nil
 	}
 	// Right now, we support extensions only on CoreOS nodes
-	if dn.OperatingSystem != MachineConfigDaemonOSRHCOS && dn.OperatingSystem != MachineConfigDaemonOSFCOS {
+	if !dn.os.IsCoreOSVariant() {
 		return fmt.Errorf("extensions is not supported on non-CoreOS nodes ")
 	}
 
 	// Validate extensions allowlist on RHCOS nodes
-	if err := validateExtensions(newConfig.Spec.Extensions); err != nil && dn.OperatingSystem == MachineConfigDaemonOSRHCOS {
+	if err := validateExtensions(newConfig.Spec.Extensions); err != nil && dn.os.IsRHCOS() {
 		return err
 	}
 
 	args := dn.generateExtensionsArgs(oldConfig, newConfig)
 	glog.Infof("Applying extensions : %+q", args)
-	if err := exec.Command("rpm-ostree", args...).Run(); err != nil {
-		return fmt.Errorf("failed to execute rpm-ostree %+q : %v", args, err)
-	}
-
-	return nil
+	_, err := runGetOut("rpm-ostree", args...)
 
+	return err
 }
 
 // switchKernel updates kernel on host with the kernelType specified in MachineConfig.
@@ -977,50 +1150,45 @@ func (dn *Daemon) switchKernel(oldConfig, newConfig *mcfgv1.MachineConfig) error
 	}
 
 	// We support Kernel update only on RHCOS nodes
-	if dn.OperatingSystem != MachineConfigDaemonOSRHCOS {
+	if !dn.os.IsRHCOS() {
 		return fmt.Errorf("updating kernel on non-RHCOS nodes is not supported")
 	}
 
 	defaultKernel := []string{"kernel", "kernel-core", "kernel-modules", "kernel-modules-extra"}
 	realtimeKernel := []string{"kernel-rt-core", "kernel-rt-modules", "kernel-rt-modules-extra", "kernel-rt-kvm"}
-	var args []string
 
 	dn.logSystem("Initiating switch from kernel %s to %s", canonicalizeKernelType(oldConfig.Spec.KernelType), canonicalizeKernelType(newConfig.Spec.KernelType))
 
 	if canonicalizeKernelType(oldConfig.Spec.KernelType) == ctrlcommon.KernelTypeRealtime && canonicalizeKernelType(newConfig.Spec.KernelType) == ctrlcommon.KernelTypeDefault {
-		args = []string{"override", "reset"}
+		args := []string{"override", "reset"}
 		args = append(args, defaultKernel...)
 		for _, pkg := range realtimeKernel {
 			args = append(args, "--uninstall", pkg)
 		}
 		dn.logSystem("Switching to kernelType=%s, invoking rpm-ostree %+q", newConfig.Spec.KernelType, args)
-		if err := exec.Command("rpm-ostree", args...).Run(); err != nil {
-			return fmt.Errorf("failed to execute rpm-ostree %+q : %v", args, err)
-		}
-		return nil
+		_, err := runGetOut("rpm-ostree", args...)
+		return err
 	}
 
 	if canonicalizeKernelType(oldConfig.Spec.KernelType) == ctrlcommon.KernelTypeDefault && canonicalizeKernelType(newConfig.Spec.KernelType) == ctrlcommon.KernelTypeRealtime {
 		// Switch to RT kernel
-		args = []string{"override", "remove"}
+		args := []string{"override", "remove"}
 		args = append(args, defaultKernel...)
 		for _, pkg := range realtimeKernel {
 			args = append(args, "--install", pkg)
 		}
 
 		dn.logSystem("Switching to kernelType=%s, invoking rpm-ostree %+q", newConfig.Spec.KernelType, args)
-		if err := exec.Command("rpm-ostree", args...).Run(); err != nil {
-			return fmt.Errorf("failed to execute rpm-ostree %+q : %v", args, err)
-		}
-		return nil
+		_, err := runGetOut("rpm-ostree", args...)
+		return err
 	}
 
 	if canonicalizeKernelType(oldConfig.Spec.KernelType) == ctrlcommon.KernelTypeRealtime && canonicalizeKernelType(newConfig.Spec.KernelType) == ctrlcommon.KernelTypeRealtime {
 		if oldConfig.Spec.OSImageURL != newConfig.Spec.OSImageURL {
+			args := []string{"update"}
 			dn.logSystem("Updating rt-kernel packages on host: %+q", args)
-			if err := exec.Command("rpm-ostree", "update").Run(); err != nil {
-				return fmt.Errorf("failed to execute rpm-ostree %+q : %v", args, err)
-			}
+			_, err := runGetOut("rpm-ostree", args...)
+			return err
 		}
 	}
 
@@ -1124,11 +1292,6 @@ func (dn *Daemon) deleteStaleData(oldIgnConfig, newIgnConfig *ign3types.Config)
 		newFileSet[f.Path] = struct{}{}
 	}
 
-	operatingSystem, err := GetHostRunningOS()
-	if err != nil {
-		return errors.Wrapf(err, "checking operating system")
-	}
-
 	for _, f := range oldIgnConfig.Storage.Files {
 		if _, ok := newFileSet[f.Path]; ok {
 			continue
@@ -1146,7 +1309,7 @@ func (dn *Daemon) deleteStaleData(oldIgnConfig, newIgnConfig *ign3types.Config)
 			if _, err := exec.Command("rpm", "-qf", f.Path).CombinedOutput(); err == nil {
 				// File is owned by an rpm
 				restore = true
-			} else if strings.HasPrefix(f.Path, "/etc") && (operatingSystem == MachineConfigDaemonOSRHCOS || operatingSystem == MachineConfigDaemonOSFCOS) {
+			} else if strings.HasPrefix(f.Path, "/etc") && dn.os.IsCoreOSVariant() {
 				if _, err := os.Stat("/usr" + f.Path); err != nil {
 					if !os.IsNotExist(err) {
 						return err
@@ -1230,8 +1393,12 @@ func (dn *Daemon) deleteStaleData(oldIgnConfig, newIgnConfig *ign3types.Config)
 		}
 		path := filepath.Join(pathSystemd, u.Name)
 		if _, ok := newUnitSet[path]; !ok {
-			if err := dn.disableUnit(u); err != nil {
-				glog.Warningf("Unable to disable %s: %s", u.Name, err)
+			// since the unit doesn't exist anymore within the MachineConfig,
+			// look to restore defaults here, so that symlinks are removed first
+			// if the system has the service disabled
+			// writeUnits() will catch units that still have references in other MCs
+			if err := dn.presetUnit(u); err != nil {
+				glog.Infof("Did not restore preset for %s (may not exist): %s", u.Name, err)
 			}
 			if _, err := os.Stat(noOrigFileStampName(path)); err == nil {
 				if err := os.Remove(noOrigFileStampName(path)); err != nil {
@@ -1261,46 +1428,76 @@ func (dn *Daemon) deleteStaleData(oldIgnConfig, newIgnConfig *ign3types.Config)
 	return nil
 }
 
-// enableUnit enables a systemd unit via symlink
-func (dn *Daemon) enableUnit(unit ign3types.Unit) error {
-	// The link location
-	wantsPath := filepath.Join(wantsPathSystemd, unit.Name)
-	// sanity check that we don't return an error when the link already exists
-	if _, err := os.Stat(wantsPath); err == nil {
-		glog.Infof("%s already exists. Not making a new symlink", wantsPath)
-		return nil
-	}
-	// The originating file to link
-	servicePath := filepath.Join(pathSystemd, unit.Name)
-	err := renameio.Symlink(servicePath, wantsPath)
+// enableUnits enables a set of systemd units via systemctl, if any fail all fails.
+func (dn *Daemon) enableUnits(units []string) error {
+	args := append([]string{"enable"}, units...)
+	stdouterr, err := exec.Command("systemctl", args...).CombinedOutput()
 	if err != nil {
-		return err
+		if !dn.os.IsLikeTraditionalRHEL7() {
+			return fmt.Errorf("error enabling units: %s", stdouterr)
+		}
+		// In RHEL7, the systemd version is too low, so it is unable to handle broken
+		// symlinks during enable. Do a best-effort removal of potentially broken
+		// hard coded symlinks and try again.
+		// See: https://bugzilla.redhat.com/show_bug.cgi?id=1913536
+		wantsPathSystemd := "/etc/systemd/system/multi-user.target.wants/"
+		for _, unit := range units {
+			unitLinkPath := filepath.Join(wantsPathSystemd, unit)
+			fi, fiErr := os.Lstat(unitLinkPath)
+			if fiErr != nil {
+				if !os.IsNotExist(fiErr) {
+					return fmt.Errorf("error trying to enable unit, fallback failed with %s (original error %s)",
+						fiErr, stdouterr)
+				}
+				continue
+			}
+			if fi.Mode()&os.ModeSymlink == 0 {
+				return fmt.Errorf("error trying to enable unit, a non-symlink file exists at %s (original error %s)",
+					unitLinkPath, stdouterr)
+			}
+			if _, evalErr := filepath.EvalSymlinks(unitLinkPath); evalErr != nil {
+				// this is a broken symlink, remove
+				if rmErr := os.Remove(unitLinkPath); rmErr != nil {
+					return fmt.Errorf("error trying to enable unit, cannot remove broken symlink: %s (original error %s)",
+						rmErr, stdouterr)
+				}
+			}
+		}
+		stdouterr, err := exec.Command("systemctl", args...).CombinedOutput()
+		if err != nil {
+			return fmt.Errorf("error enabling units: %s", stdouterr)
+		}
 	}
-	glog.Infof("Enabled %s", unit.Name)
-	glog.V(2).Infof("Symlinked %s to %s", servicePath, wantsPath)
+	glog.Infof("Enabled systemd units: %v", units)
 	return nil
 }
 
-// disableUnit disables a systemd unit via symlink removal
-func (dn *Daemon) disableUnit(unit ign3types.Unit) error {
-	// The link location
-	wantsPath := filepath.Join(wantsPathSystemd, unit.Name)
-	// sanity check so we don't return an error when the unit was already disabled
-	if _, err := os.Stat(wantsPath); err != nil {
-		glog.Infof("%s was not present. No need to remove", wantsPath)
-		return nil
+// disableUnits disables a set of systemd units via systemctl, if any fail all fails.
+func (dn *Daemon) disableUnits(units []string) error {
+	args := append([]string{"disable"}, units...)
+	stdouterr, err := exec.Command("systemctl", args...).CombinedOutput()
+	if err != nil {
+		return fmt.Errorf("error disabling unit: %s", stdouterr)
 	}
-	glog.V(2).Infof("Disabling unit at %s", wantsPath)
+	glog.Infof("Disabled systemd units %v", units)
+	return nil
+}
 
-	return os.Remove(wantsPath)
+// presetUnit resets a systemd unit to its preset via systemctl
+func (dn *Daemon) presetUnit(unit ign3types.Unit) error {
+	args := []string{"preset", unit.Name}
+	stdouterr, err := exec.Command("systemctl", args...).CombinedOutput()
+	if err != nil {
+		return fmt.Errorf("error running preset on unit: %s", stdouterr)
+	}
+	glog.Infof("Preset systemd unit %s", unit.Name)
+	return nil
 }
 
 // writeUnits writes the systemd units to disk
 func (dn *Daemon) writeUnits(units []ign3types.Unit) error {
-	operatingSystem, err := GetHostRunningOS()
-	if err != nil {
-		return errors.Wrapf(err, "checking operating system")
-	}
+	var enabledUnits []string
+	var disabledUnits []string
 	for _, u := range units {
 		// write the dropin to disk
 		for i := range u.Dropins {
@@ -1322,7 +1519,7 @@ func (dn *Daemon) writeUnits(units []ign3types.Unit) error {
 
 			glog.Infof("Writing systemd unit dropin %q", u.Dropins[i].Name)
 			if _, err := os.Stat("/usr" + dpath); err == nil &&
-				(operatingSystem == MachineConfigDaemonOSRHCOS || operatingSystem == MachineConfigDaemonOSFCOS) {
+				dn.os.IsCoreOSVariant() {
 				if err := createOrigFile("/usr"+dpath, dpath); err != nil {
 					return err
 				}
@@ -1356,7 +1553,7 @@ func (dn *Daemon) writeUnits(units []ign3types.Unit) error {
 		if u.Contents != nil && *u.Contents != "" {
 			glog.Infof("Writing systemd unit %q", u.Name)
 			if _, err := os.Stat("/usr" + fpath); err == nil &&
-				(operatingSystem == MachineConfigDaemonOSRHCOS || operatingSystem == MachineConfigDaemonOSFCOS) {
+				dn.os.IsCoreOSVariant() {
 				if err := createOrigFile("/usr"+fpath, fpath); err != nil {
 					return err
 				}
@@ -1370,23 +1567,40 @@ func (dn *Daemon) writeUnits(units []ign3types.Unit) error {
 		}
 
 		// if the unit doesn't note if it should be enabled or disabled then
-		// skip all linking.
-		// if the unit should be enabled, then enable it.
-		// otherwise the unit is disabled. run disableUnit to ensure the unit is
-		// disabled. even if the unit wasn't previously enabled the result will
-		// be fine as disableUnit is idempotent.
+		// honour system presets. This to account for an edge case where you
+		// deleted a MachineConfig that enabled/disabled the unit to revert,
+		// but the unit itself is referenced in other MCs. deleteStaleData() will
+		// catch fully deleted units.
+		// if the unit should be enabled/disabled, then enable/disable it.
+		// this is a no-op if the system wasn't change this iteration
+		// Also, enable and disable as one command, as if any operation fails
+		// we'd bubble up the error anyways, and we save a lot of time doing this.
+		// Presets must be done individually as we don't consider a failed preset
+		// as an error, but it would cause other presets that would have succeeded
+		// to not go through.
+
 		if u.Enabled != nil {
 			if *u.Enabled {
-				if err := dn.enableUnit(u); err != nil {
-					return err
-				}
-				glog.V(2).Infof("Enabled systemd unit %q", u.Name)
+				enabledUnits = append(enabledUnits, u.Name)
 			} else {
-				if err := dn.disableUnit(u); err != nil {
-					return err
-				}
-				glog.V(2).Infof("Disabled systemd unit %q", u.Name)
+				disabledUnits = append(disabledUnits, u.Name)
 			}
+		} else {
+			if err := dn.presetUnit(u); err != nil {
+				// Don't fail here, since a unit may have a dropin referencing a nonexisting actual unit
+				glog.Infof("Could not reset unit preset for %s, skipping. (Error msg: %v)", u.Name, err)
+			}
+		}
+	}
+
+	if len(enabledUnits) > 0 {
+		if err := dn.enableUnits(enabledUnits); err != nil {
+			return err
+		}
+	}
+	if len(disabledUnits) > 0 {
+		if err := dn.disableUnits(disabledUnits); err != nil {
+			return err
 		}
 	}
 	return nil
@@ -1398,9 +1612,20 @@ func (dn *Daemon) writeFiles(files []ign3types.File) error {
 	for _, file := range files {
 		glog.Infof("Writing file %q", file.Path)
 
-		contents, err := dataurl.DecodeString(*file.Contents.Source)
-		if err != nil {
-			return err
+		// We don't support appends in the file section, so instead of waiting to fail validation,
+		// let's explicitly fail here.
+		if len(file.Append) > 0 {
+			return fmt.Errorf("found an append section when writing files. Append is not supported")
+		}
+
+		// To allow writing of "empty" files we'll allow source to be nil
+		contents := &dataurl.DataURL{}
+		if file.Contents.Source != nil {
+			var err error
+			contents, err = dataurl.DecodeString(*file.Contents.Source)
+			if err != nil {
+				return err
+			}
 		}
 		mode := defaultFilePermissions
 		if file.Mode != nil {
@@ -1541,7 +1766,7 @@ func (dn *Daemon) updateSSHKeys(newUsers []ign3types.PasswdUser) error {
 
 // updateOS updates the system OS to the one specified in newConfig
 func (dn *Daemon) updateOS(config *mcfgv1.MachineConfig, osImageContentDir string) error {
-	if dn.OperatingSystem != MachineConfigDaemonOSRHCOS && dn.OperatingSystem != MachineConfigDaemonOSFCOS {
+	if !dn.os.IsCoreOSVariant() {
 		glog.V(2).Info("Updating of non-CoreOS nodes are not supported")
 		return nil
 	}
